//
//  $Id: mover.cc,v 1.1 1997/09/30 12:00:00 jochen Rel $ 
//  Copyright (C) 1997  Jochen Metzinger<1metzing@informatik.uni-hamburg.de>
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software Foundation,
//  Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//

#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <vga.h>
#include <vgagl4.h>
#include <time.h>	/* for time() */

#define	VGAMODE	G640x480x16

#define EGA_BLACK		 0
#define EGA_BLUE		 1
#define EGA_GREEN		 2
#define EGA_CYAN		 3
#define EGA_RED			 4
#define EGA_MAGENTA		 5
#define EGA_BROWN		 6
#define EGA_LIGHTGRAY		 7
#define EGA_DARKGRAY		 8
#define EGA_LIGHTBLUE		 9
#define EGA_LIGHTGREEN		10
#define EGA_LIGHTCYAN		11
#define EGA_LIGHTRED		12
#define EGA_LIGHTMAGENTA	13
#define EGA_YELLOW		14
#define EGA_WHITE		15

#define BACKGROUND	EGA_BLACK
#define MARGIN		EGA_BLUE
#define FEED		EGA_LIGHTGREEN
#define MOVER		EGA_LIGHTGRAY


class Mover {
	int x, y, dir;
	int energy, age;
	int gen[6], p[6];
public:
	Mover();
	~Mover();
	void random();
	bool alive() {return (energy > 0);}
	bool able_to_split() {return (age > 800) && (energy >= 1000);}
	void move();
	void split(Mover &child);
protected:
	void change_dir();
	void init_dir(bool change=false);
	static int at(int x, int y);
private:
	static const int dx[6];
	static const int dy[6];
};


const int Mover::dx[6] = { 0,  2,  2,  0, -2, -2};
const int Mover::dy[6] = { 2,  1, -1, -2, -1,  1};


int
Mover::at(int x, int y)
{
	int i, j, c, food;

	food = 0;
	for (j=-1; j<=1; j++) {
		for (i=-1; i<=1; i++) {
			c = gl_getpixel(x+i, y+j);
			if (c == BACKGROUND) {
				continue;
			} else if (c == FEED) {
				food++;
			} else {
				return -1;
			}
		}
	}
	return food;
}


Mover::Mover()
:energy(0), age(-1)
{
}


void
Mover::random()
{
	int i, food;

	age = 0;
	energy = 500;
	do {
		x = rand() % WIDTH;
		y = rand() % HEIGHT;
		food = at(x,y);
	} while (food < 0);
	energy += 40*food;
	for (i=0; i<6; i++) {
		gen[i] = (rand() % 5) - 2;
	}
	init_dir();
	gl_fillbox(x-1, y-1, 3, 3, MOVER);
}


Mover::~Mover()
{
	gl_fillbox(x-1, y-1, 3, 3, BACKGROUND);
}


void
Mover::init_dir(bool change)
{
	int i;
	double pp[6], sum;

	if (change) {
		i = rand() % 6;
		if (rand() % 2)
			gen[i]++;
		else
			gen[i]--;
	}
	pp[0] = ldexp(1.0, gen[0]); /* 2 ** gen[0] */
	for (i=1; i<6; i++) {
		pp[i] = pp[i-1] + ldexp(1.0, gen[i]);
	}
	sum = pp[5];
	for (i=1; i<6; i++) {
		p[i] = (int)((double)RAND_MAX*pp[i]/sum);
	}
}


void
Mover::change_dir()
{
	int dd, r;

	do {
		r = rand();
		for (dd=0; dd<6; dd++) {
			if (r <= p[dd]) {
				break;
			}
		}
	} while (dd == 6);
	dir = (dir + dd) % 6;
}


void
Mover::move()
{
	int x1, y1, food;

	if (! alive()) return;
	energy--;
	age++;
	change_dir();
	gl_fillbox(x-1, y-1, 3, 3, BACKGROUND);
	if (! alive()) return;
	x1 = x + dx[dir];
	y1 = y + dy[dir];
	food = at(x1, y1);
	if (food >= 0) {
		energy += 40*food;
		if (energy > 1500) energy = 1500;
		x = x1;
		y = y1;
	}
	gl_fillbox(x-1, y-1, 3, 3, MOVER);
}


void
Mover::split(Mover &child)
{
	int i, food;

	// setup child
	child.x = x + 2*dx[dir];
	child.y = y + 2*dy[dir];
	child.dir = dir;
	child.energy = 0;
	child.age = 0;
	for (i=0; i<6; i++) child.gen[i] = gen[i];
	child.init_dir(true);
	// test if place ok
	food = at(child.x, child.y);
	if (food >= 0) {
		// setup child & me
		dir = (dir + 3) % 6; 
		energy /=2;
		init_dir(true);
		child.age = age = 0;
		child.energy = energy + 40*food;
		if (child.energy > 1500) child.energy = 1500;
	}
	// draw it
	gl_fillbox(child.x-1, child.y-1, 3, 3, MOVER);
}


void
add_feed(int n=1)
{
	int x, y;

	while (n-- > 0) {
		x = 1 + rand()%(WIDTH-2);
		y = 1 + rand()%(HEIGHT-2);
		if (BACKGROUND == gl_getpixel(x, y)) {
			gl_setpixel(x, y, FEED);
		}
	}
}


#define MAX_MOVER 500

Mover MoverPool[MAX_MOVER];
int nr_biest;


int
new_index()
{
	int i;

	for (i=nr_biest-1; i>=0; i--) {
		if (! MoverPool[i].alive()) {
			return i;
		}
	}
	nr_biest++;
	assert(nr_biest <= MAX_MOVER);
	return nr_biest-1;
}


void
init()
{
	int i;

	// frame whole screen
	gl_line(0, 0, WIDTH-1, 0, MARGIN);
	gl_line(WIDTH-1, 0, WIDTH-1, HEIGHT-1, MARGIN);
	gl_line(WIDTH-1, HEIGHT-1, 0, HEIGHT-1, MARGIN);
	gl_line(0, HEIGHT-1, 0, 0, MARGIN);
	// init rand()
	srand(time(NULL));
	// setup biests
	nr_biest = 10;
	for (i=0; i<nr_biest; i++) {
		MoverPool[i].random();
	}
	// setup feed
	add_feed(WIDTH*HEIGHT/40);
}


int
main()
{
	int i, j, time;
	bool leben;

	if (vga_init() != 0)
		exit(1);
	if (vga_hasmode(VGAMODE) != 1) {
		fprintf(stderr, "vga mode %s not available",
			vga_getmodename(VGAMODE));
		exit(2);
	}
	vga_setmode(VGAMODE);
	gl4_nolockvc();
	gl4_setcontextvga(VGAMODE);
	init();
	time = 0;
	do {
		// step
		leben = false;
		for (i=0; i<nr_biest; i++) {
			if (MoverPool[i].alive()) {
				leben = true;
			}
			MoverPool[i].move();
			if (MoverPool[i].able_to_split()) {
				j = new_index();
				MoverPool[i].split(MoverPool[j]);
			}
		}
		add_feed(1);
		// update simple clock
		time++;
		gl_setpixel((time+0)%WIDTH, 0, EGA_WHITE);
		gl_setpixel((time-1)%WIDTH, 0, MARGIN);
	} while (leben && vga_getkey() == 0);
	vga_setmode(TEXT);
	return 0;
}
